# Exemple d'utilisation de matplotlib

# On utilise numpy.random.rand() pour generer des points mais en realite
# il faudra programmer son propre generateur aleatoire sans utiliser numpy.random

import sys
Windows = 'win' in sys.platform
import matplotlib.pyplot as plt
if not(Windows):
    from matplotlib.pyplot import ion
    ion()

def histogramme(l, titre='histogramme'):
    plt.figure()
    plt.title(titre)
    plt.hist(l)
    plt.show()
    return(plt)




import sys
import os
cwd = os.getcwd()
sys.path.insert(0, cwd+os.sep+".."+os.sep)

import histogrammes as hist

from corriges.gaussien import *

from corriges.melanges import *

import scipy

import pandas

input_file = ".." + os.sep + "PearsonLee.txt"

x = pandas.read_csv(input_file, skiprows=0, sep=" ", header=25)

noms_col = list(x.keys()) # Noms de colonnes

stature = scipy.array(x["parent"])

n = len(stature) # taille d'échantillon

hist.histogramme(stature, "histogramme des statures")
# Loi de melange gaussien pour estimer la variance

# theta = [[65., 5.], [66., 5.]]  # Valeur initiale des parametres
# Donne essentiellement une seule composante

theta = [[60., 5.], [70., 5.]]  # Valeur initiale des parametres

theta = [[60., 5.], [65., 5.], [70., 5.]]  # Valeur initiale des parametres

# Hyperparametres (loi a priori normale inverse-chi2)
# On prend ici les memes hyperparametres pour les deux gaussiennes
hyper_param = [0., 1., 2, 1.]

nit = 4000  # nombre d'iterations

K = len(theta)  # nombre d'etats du melange

## 2 etats
# Echantillonneur de Gibbs
mcmc2 = gibbs(nit, stature, theta, densite_norm, x_sim_norm,
              rpost_norm_invchi2, hyper_param)

# En sortie : labels des individus, loi predictive,
# liste des proportions estimees au cours des iterations
# liste des parametres d'emission estimes au cours des iterations
# tri permettant de compenser les permutations de labels

z, xpred, tri_prop, lambdap, tri = mcmc2

from matplotlib.pyplot import figure
import matplotlib.pyplot as plt

burn = 200

# Loi predictive a posteriori

fig = figure(figsize=(13, 7))
fig.add_subplot(1,2,1)
plt.hist(stature)
plt.title("histogramme des donnees")
fig.add_subplot(1,2,2)
plt.hist(xpred[burn:,0])
plt.title("histogramme de la loi predictive a posteriori")

# Proportions du melange

tri_prop[burn:,1].mean()

# Parametres d'emission pour l'etat 0

lambdap = scipy.array(lambdap)

fig = figure(figsize=(13, 7))

fig.add_subplot(1,2,1)

plt.hist(lambdap[burn:,0,0])

plt.title("Esperance a posteriori etat 0")

fig.add_subplot(1,2,2)

plt.hist(lambdap[burn:,0,1])

plt.title("Ecart-type a posteriori etat 0")


# Parametres d'emission pour l'etat 1

lambdap = scipy.array(lambdap)

fig = figure(figsize=(13, 7))

fig.add_subplot(1,2,1)

plt.hist(lambdap[burn:,1,0])

plt.title("Esperance a posteriori etat 1")

fig.add_subplot(1,2,2)

plt.hist(lambdap[burn:,1,1])

plt.title("Ecart-type a posteriori etat 1")

## 3 etats
theta = [[60., 5.], [65., 5.], [70., 5.]]  # Valeur initiale des parametres

K = len(theta)  # nombre d'etats du melange

nit = 30
burn = 10
# Echantillonneur de Gibbs
mcmc3 = gibbs(nit, stature, theta, densite_norm, x_sim_norm,
              rpost_norm_invchi2, hyper_param)

z, xpred, tri_prop, lambdap, tri = mcmc3

# Loi predictive a posteriori

fig = figure(figsize=(13, 7))
fig.add_subplot(1,2,1)
plt.hist(stature)
plt.title("histogramme des donnees")
fig.add_subplot(1,2,2)
plt.hist(xpred[burn:,0])
plt.title("histogramme de la loi predictive a posteriori")



"""Analyse bayesienne de lois normales (cas univarie)"""
import sys
import os
cwd = os.getcwd()
sys.path.insert(0, cwd+os.sep+".."+os.sep)

import scipy

from numpy.random import standard_normal
from scipy.stats import norm
from scipy.special import gamma, gammaln

def x_sim_norm(n, theta):
    """
    Simulation d'une loi normale de parametre theta
    
    :Parameters:
        `n (int)` - taille d'echantillon
        `theta (list)` - parametres

    :Returns:
        Realisation d'une variable aleatoire de loi normale
    
    :Remark:
        theta[0] : esperance de x
        theta[1] : variance de x
    """
    # Objet representant la loi normale
    loi_normale = norm(theta[0], scipy.sqrt(theta[1]))
    return(list(loi_normale.rvs(n)))

def densite_norm(x, theta):    
    """
    Densite d'une loi normale de parametre theta
    
    :Parameters:
        `x (iterable)` - valeurs en laquelles calculer la densite
        `theta (list)` - parametres

    :Returns:
        Densite en x d'une variable aleatoire de loi normale
    
    :Remark:
        theta[0] : esperance de x
        theta[1] : variance de x
    """
    loi_normale = norm(theta[0], scipy.sqrt(theta[1]))
    nnorm = lambda v: loi_normale.pdf(v)
    return(list(map(nnorm, x)))


def densite_inv_chi_deux(x, nu, s2):
    """
    Densite d'une loi inverse chi deux
    de parametres nu et s2

    :Parameters:
        `x (iterable)` - valeurs en laquelles calculer la densite
        `nu (float)` -  degres de liberte dans la loi chi deux
        `nu (float)` - parametre d'echelle

    :Returns:
        Densite en x d'une variable aleatoire de loi inverse chi deux

    """
    nf = float(nu)
    def densite(v):
        if v <= 0:
            return(0)
        else:
            s = nf * s2 / v
            # calcule la log-densite
            d = (nf/2 + 1) * scipy.log(s) -s / 2. - scipy.log(nf) - scipy.log(s2) - (nf/2) * scipy.log(2)
            d -= gammaln(nf/2.)
        return(scipy.exp(d))
    return(list(map(densite, x)))

def post_norm_invchi2(x, hyper_param):
    """
    Parametres de la loi a posteriori pour une loi normale
    univariee avec une loi a priori normale/inverse chi deux.
    
    :Parameters:
        `x (list)` - liste des donnees (format type scipy.array)
        `hyper_param (list)` - parametres de la loi a priori (hyperparametres)

    :Returns:
        Parametres de la loi a posteriori
    
    :Remark:
        hyper_param[0] : esperance a priori
        hyper_param[1] : composante multiplicative de la variance a priori 
                        de theta[0]
        hyper_param[2] : degres de liberte dans la loi chi deux / inverse 
                            chi deux
        hyper_param[3] : parametre d'echelle dans la loi inverse chi deux            
        Par convention, on considere que l'esperance est fixee si hyper_param[1] == 0
    """
    m0, rho2, nu0, sz2 = hyper_param
    n = x.shape[0] # taille d'echantillon
    # Loi a posteriori des 4 hyperparametres
    if rho2 > 0:
        xs = float(x.sum())
        rho2_post = 1. / (n + 1./rho2)
        m_post = rho2_post * (m0/rho2 + xs)
        xv = x.var()
    else:
        xs = m0
        rho2_post = 0
        m_post = m0
        xv = (x - xs)**2
        xv = float(xv.mean())
    nu_post = n + nu0
    s2_post = (n * xv + nu0 * sz2 + (m0-xs/n)**2 * n / (1 + n*rho2)) 
    s2_post /= nu_post
    s2_sim = nu_post * s2_post / scipy.stats.chi2.rvs(nu_post, size=1)
    m_sim = (s2_post * rho2_post)**0.5 * standard_normal(1) + m_post
    return((m_post, rho2_post, nu_post, s2_post))


def dpost_norm_invchi2(theta, x, hyper_param):
    """
    Densite de la loi a posteriori pour une loi normale
    univariee avec une loi a priori normale/inverse chi deux

    :Parameters:
        `theta (list)` - parametres
        `x (list)` - liste des donnees (format type scipy.array)
        `hyper_param (list)` - parametres de la loi a priori (hyperparametres)

    :Returns:
        Densite de la loi a posteriori (esperance, ecart-type) en theta

    :Remark:
        theta[0] : esperance de x
        theta[1] : variance de x
        hyper_param[0] : esperance a priori
        hyper_param[1] : composante multiplicative de la variance a priori 
                        de theta[0]
        hyper_param[2] : degres de liberte dans la loi chi deux / inverse 
                            chi deux
        hyper_param[3] : parametre d'echelle dans la loi inverse chi deux            
    """
    m_post, rho2_post, nu_post, s2_post = post_norm_invchi2(x, hyper_param)
    # densite marginale a posteriori de la variance
    ds2 = densite_inv_chi_deux(theta[1], nu_post, s2_post)[0]
    # densite conditionnelle a posteriori de l'esperance sachant la variance
    dm = densite_norm(theta[0], [m_post, rho2_post * s2_post])[0]
    return(dm * ds2)


def rpost_invchi2(x, hyper_param):
    """
    Simule la loi a posteriori pour une loi normale
    univariee avec une loi a priori normale/inverse chi deux

    :Parameters:
        `x (list)` - liste des donnees (format type scipy.array)
        `hyper_param (list)` - parametres de la loi a priori (hyperparametres)

    :Returns:
        Realisation de theta suivant la loi a posteriori (esperance, variance)

    :Remark:
        hyper_param[0] : degres de liberte dans la loi chi deux / inverse
                            chi deux
        hyper_param[1] : parametre d'echelle dans la loi inverse chi deux
    """
    m_post, rho2_post, nu_post, s2_post = post_norm_invchi2(x, hyper_param)
    s2_sim = nu_post * s2_post / scipy.stats.chi2.rvs(nu_post, size=1)
    return ([s2_sim[0]])

def rpost_norm_invchi2(x, hyper_param):
    """
    Simule la loi a posteriori pour une loi normale
    univariee avec une loi a priori normale/inverse chi deux

    :Parameters:
        `x (list)` - liste des donnees (format type scipy.array)
        `hyper_param (list)` - parametres de la loi a priori (hyperparametres)

    :Returns:
        Realisation de theta suivant la loi a posteriori (esperance, variance)

    :Remark:
        hyper_param[0] : esperance a priori
        hyper_param[1] : composante multiplicative de la variance a priori
                        de theta[0]
        hyper_param[2] : degres de liberte dans la loi chi deux / inverse
                            chi deux
        hyper_param[3] : parametre d'echelle dans la loi inverse chi deux
    """
    m_post, rho2_post, nu_post, s2_post = post_norm_invchi2(x, hyper_param)
    s2_sim = rpost_invchi2(x, hyper_param)
    m_sim = (s2_post * rho2_post) ** 0.5 * standard_normal(1) + m_post
    return ([m_sim[0], s2_sim[0]])



# -*- coding: utf-8 -*-
"""
Created on Thu Jun 29 11:39:00 2017

@author: durand

ACP probabiliste :
Calcule l'inverse de la matrice de variance-covariance de la loi de W|X,Z, ...
Modele : X = mu + WZ + sigma epsilon
reecrit X = mu + UW  + sigma epsilon
V = U * U.T + sigma**2 I
"""

import scipy, scipy.linalg

def CalculU(Z,D,sigma2=0):
    """
    Calcule U
    """
    n, M = Z.shape
    U = scipy.matrix(scipy.zeros((n*D,D*M)))
    for i in range(n):
        for j in range(D):
            # print(i*D+j)
            # print(range((j*M),(j*M+M)))
            U[i*D+j,(j*M):(j*M+M)] = Z[i,:]
    return(U)
    
#def InvV(Z,D,sigma2=0):
#    """
#    Calcule V^-1    
#    """
#    
    
    
# Exemple
n = 3
M = 4
D = 5
vals = scipy.randn(n*M)
Z = scipy.matrix(vals)
Z = Z.reshape(n,M)
U = CalculU(Z,D,0)
V = U*U.T
VM1 = scipy.linalg.inv(V)
print(VM1)
Sigma = Z*Z.T
print(scipy.linalg.inv(Sigma))
UTVM1 = U.T * VM1
print(UTVM1)
UTVM1U = UTVM1 * U
print(UTVM1U)
UTVM1U[0:M,0:M]
UTVM1U[M:2*M,M:2*M]
Z.T * scipy.linalg.inv(Sigma) * Z

# On remarque que la matrice de variance-covariance de la loi de W|X,Z est
# diagonale par blocs, les blocs diagonaux etant egaux a 
# Z.T * (Z * Z.T + sigma^2 * I)^{-1} * Z


import sys
import os
cwd = os.getcwd()
sys.path.insert(0, cwd+os.sep+".."+os.sep)

import histogrammes as hist

from corriges.gaussien import *

import scipy

import pandas

input_file = ".." + os.sep + "SBF250.xls"

x = pandas.read_excel(input_file)

cac = x["FRCAC40"]

f_cac = [c for c in cac if not(pandas.isnull(c))]
f_cac = scipy.array(f_cac)

n = len(f_cac) # taille d'echantillon

hist.histogramme(f_cac, "histogramme des rentabilites du CAC40")

n_post = 1000 # nombre de tirages suivant la loi a posteriori
# Simulation loi a posteriori normale inverse-chi2
hyper_param = [0., 1., 2, 1.]
p = list(map(lambda r:rpost_norm_invchi2(f_cac,
               hyper_param), range(n_post)))
p = scipy.array(p)

hist.histogramme(p[:,0], "histogramme de l'esperance a posteriori")
hist.histogramme(p[:,1], "histogramme de l'ecart-type a posteriori")


"""Analyse bayesienne de melanges de lois"""
import sys
import os
import scipy, numpy
cwd = os.getcwd()
sys.path.insert(0, cwd+os.sep+".."+os.sep)


def z_posteriori(x, theta, prob_cond):
    """
    Calcul de la loi a posteriori des etats
    sachant les observations et les parametres d'emission
    
    :Parameters:
        `x (list)` - liste des donnees
        `theta (list)` - liste des parametres d'emission
        `prob_cond (lambda x, theta:)` - fonction de densite conditionnelle
            lambda x: prob_cond(x, theta) = p(x|theta)

    :Returns:
        Retourne un tableau de probas conditionnelles : 
        Ligne i = observations x[i]
        Colonne k = etats k            
    """
    K = len(theta)  # Nombre d'etats
    n = len(x)  # Taille d'echantillon
    post = scipy.zeros((n, K))
    # Matrice des probas conditionnelles : 
    # Lignes = observations x
    # Colonnes = etats
    for k in range(K):
        post[:, k] = prob_cond(x, theta[k])
    # Normalisation des lignes par leur somme
    vect = post.sum(1)
    post = post / vect[:, None]
    return(post)


def gibbs(nit, x, lambda0, prob_cond, x_sim, rtheta_post, hyper_param):
    """
    Calcul de la loi a posteriori des parametres d'emission
    sachant les observations
    
    :Parameters:
        `nit (int)` - nombre d'iterations
        `x (list)` - donnees (liste ou format type scipy.array)
        `lambda0 (list)` - liste des parametres d'emission initiaux
        `prob_cond (lambda x, theta:)` - fonction de densite conditionnelle
            lambda x: prob_cond(x, theta) = p(x|theta)
        `x_sim (lambda n, theta:)` - fonction de simulation d'un echantillon
            de taille n suivant la loi p(x|theta)
        `rthetat_post (lambda x, hyper_param:)` - fonction de simulation
            suivant la loi a posteriori des parametres (sachant x et les hyper-
            parametres
        `hyper_param (list)` - liste des hyper-parametres

    :Returns:
        Retourne un quintuplet  (z, xpred, tri_prop, lambdap, tri) avec:
        `z` (scipy.array) - valeurs des etats simulees
            (lignes: iterations, colonnes: individus)
        `xpred` (scipy.array) - simulations suivant la loi predictive a
            posteriori (meme format que x)
        `tri_prop` (scipy.array) - frequence des etats au cours des iterations
            (lignes: iterations, colonnes: etats)
        `lambdap` (liste) - parametres simules au cours des iterations
        `tri`  (scipy.array) - permutations des etats realisees au cours
            des iterations pour conserver l'ordre des parametres
            (lignes: iterations, colonnes: etats)

    :Remark:
        - Les etats sont numerotes de 0 a len(lambda0)-1
        - x peut etre une liste de liste dans le cas multidimensionnel, dans 
        ce cas il doit etre formatte comme un scipy.array: les individus sont
        associes aux elements de la liste englobante, les dimensions a ceux
        des listes englobees.

    :SeeAlso:
        z_posteriori
        
    """
    dx = scipy.array(x)
    K = len(lambda0)  # nombre d'etats
    n = len(x)  # taille d'echantillon
    xpred = []  # tirages suivant la loi predictive
    propnaive = scipy.array([]) # 
    tri = []  # permutations des index des clusters permettant d'avoir
            # des parametres tries par ordre lexicographique (permutation 
            # des labels
    tri_prop = []  # liste des proportions estimees
    lambdac = [[]] * K  # parametres a l'iteration courante
    lambdap = []  # liste des parametres simules
    z = []  # liste des clusters simules
    sx = scipy.array([0] * K) # nombre de points dans chaque etat
    z0 = scipy.array([-1.] * n) # cluster de chaque point
    # Tirage des etats initiaux independamment suivant
    # une loi uniforme (pas de classe vide)
    while (sx.prod() == 0):
        z0 = scipy.random.choice(K,n,p=scipy.array([1.]*K) / K)
        sx = scipy.bincount(z0)
    # Gibbs sampling pour melanges
    for it in range(nit):
        if (it > 0) and (it % 500) == 0:
            print("Iteration courante : ", it+1)
        z0 = scipy.array([-1.] * n)
        zpost = z_posteriori(dx, lambda0, prob_cond)
        # On ne veut pas de syk = 0
        sx = scipy.array([0] * K) 
        while (sx.prod() == 0):
            # simulation de p(z|y,theta)
            z0 = list(map(lambda prob: scipy.random.choice(K,1,p=prob)[0], 
                          zpost))
            z0 = scipy.array(z0)
            sx = scipy.bincount(z0)
        # simulation de p(theta|x,z)        
        for k in range(K):
            lambdac[k] = rtheta_post(dx[z0==k], hyper_param)
        lambdaca = scipy.array(lambdac)
        # estimation naive des proportions du melange 
        # (pas de loi a priori de type Dirichlet ici)
        tpropnaive = sx / sum(sx)
        # Tri de lambda0 par ordre lexicographique
        # On triera a la fin seulement
        s = scipy.lexsort(lambdaca[:,::-1].transpose())
        if it == 0:
            tri = s
            tri_prop = list(tpropnaive)
        else:
            tri = scipy.vstack((tri, s))
            tri_prop =  scipy.vstack((tri_prop, tpropnaive))
        # simulation de xpred (loi predictive):
        # simulation du z de xpred
        zpred = scipy.random.choice(K, 1, p=tpropnaive)
        assert(len(zpred)==1)
        zpred = zpred[0]
        if it == 0:
            xpred = x_sim(1, lambdac[zpred])
            z = z0
        else:
            xpred = scipy.vstack((xpred, x_sim(1, lambdac[zpred])))
            z = scipy.vstack((z, z0))
        lambdap += [list(lambdac)]
    # Reordonne les etiquettes de classe (label-switching)
    for it in range(nit):
        permm = scipy.zeros((K, K), dtype=numpy.int)  # matrice de permutation des parametres
        for j in range(K):
            permm[tri[it, j], j] = 1
        perml = lambda s: tri[it, s]
        z[it, :] = list(map(perml, z[it,:]))
        # for j in range(n):
        #    z[it, j] = tri[it, z[it, j]]
        tri_prop[it] = permm.dot(tri_prop[it])
        lambdac = list(lambdap[it])
        for j in range(K):
            lambdap[it][tri[it, j]] = lambdac[j]
    return(z, xpred, tri_prop, lambdap, tri)




import sys

import os

cwd = os.getcwd()

sys.path.insert(0, cwd)

import pathlib

sys.path.insert(0, str(pathlib.Path(cwd).parent))

sys.path.insert(0, cwd+os.sep+".."+os.sep+".."+os.sep)

import histogrammes as hist

from corriges.gaussien import *

from corriges.melanges import *

import scipy

# theta[i]: parametres (esperance, ecart-type) pour le cluster i

theta = [[7., 2.**2],[-10., 3.**2]]

# Simulation du jeu de données

n1 = 200 # Nombre d'individus du cluster 1

n2 = 300 # Nombre d'individus du cluster 2

# Simulation des donnees du cluster 1

x = x_sim_norm(200, theta[0]) 

# Simulation des donnees du cluster 2

x +=  x_sim_norm(300, theta[1])


hist.histogramme(x, "histogramme du jeu de donnees simule")


z = z_posteriori(x, theta, densite_norm)

z[190:210]

nit = 1000 # Nombre d'iterations

lambda0 = [[-1., 1],[1., 1]] # Valeur initiale des parametres


# Hyperparametres (loi a priori normale inverse-chi2)

# On prend ici les memes hyperparametres pour les deux gaussiennes

hyper_param = [0., 1., 2, 1.]    

n = len(x) # taille d'echantillon 

nit = 40  # 00 # nombre d'iterations

K = len(theta) # nombre d'etats du melange

# Echantillonneur de Gibbs

mcmc = gibbs(nit, x, theta, densite_norm, x_sim_norm,
             rpost_norm_invchi2, hyper_param)


# En sortie : labels des individus, loi predictive, estimation des proportions,

# 

z, xpred, propnaive, tri_prop, lambdap, tri = mcmc

burn = 200 # burn-in, nombre de premieres observations a enlever



hist.histogramme(x)

hist.histogramme(xpred[burn:])

hist.histogramme(propnaive[burn:,1])

lambdap = scipy.array(lambdap)

hist.histogramme(lambdap[burn:,0,0])

hist.histogramme(lambdap[burn:,0,1])

hist.histogramme(lambdap[burn:,1,0])

hist.histogramme(lambdap[burn:,1,1])





